package suggestions

import (
	"fmt"
	"log"
	"os"

	tea "github.com/charmbracelet/bubbletea"

	"github.com/tgrosinger/ledger-tui/pkg/ledger/hledger"
	"github.com/tgrosinger/ledger-tui/pkg/transaction"
	"github.com/tgrosinger/ledger-tui/pkg/tui"
)

// TransactionsMsg indicates that transactions were successfully loaded and are
// now available for use by the suggestions engine.
type TransactionsMsg []transaction.Transaction

type SuggestionErrMsg tui.ErrMsg

type Model struct {
	ledgerFilename string
	transactions   []transaction.Transaction
}

func New(ledgerFilename string) Model {
	return Model{
		ledgerFilename: ledgerFilename,
	}
}

// loadTransactions returns a bubbletea command which can be executed
// asynchronously to load transactions from a specified ledger file.
func loadTransactions(filepath string) func() tea.Msg {
	return func() tea.Msg {
		log.Println("Loading ledger file at " + filepath)

		ledger := hledger.New(filepath)
		txs, err := ledger.GetTransactions()
		if err != nil {
			return SuggestionErrMsg{Err: err}
		}

		return TransactionsMsg(txs)
	}
}

func (m Model) Init() tea.Cmd {
	return loadTransactions(m.ledgerFilename)
}

func (m Model) Update(msg tea.Msg) (Model, tea.Cmd) {

	switch msg := msg.(type) {
	case TransactionsMsg:
		log.Println("Transactions loaded")
		m.transactions = msg
	case SuggestionErrMsg:
		// TODO: This error is not visible to the user anywhere
		fmt.Fprintln(os.Stderr, "Encountered an error: "+msg.Err.Error())
		log.Println("Encountered an error: " + msg.Err.Error())
		return m, tea.Quit
	}

	return m, nil
}

func (m Model) View() string {
	return "Suggestions"
}
